/*****************************************************************************/
/*                                                                           */
/*    CS 473 -- Fall, 2004                                                   */
/*    HW1 Sample Solution                                                    */
/*                                                                           */
/*    This program will read two floating point numbers from stdin, in the   */
/*    format specified in the assignment, multiply them together, and print  */
/*    the result to standard out                                             */
/*                                                                           */
/*****************************************************************************/
#include <ctype.h>
#include <stdint.h>
#include <stdio.h>

// Let's do a typedef to make my intended data type clear.  Also, I
// like to have a bool
typedef uint32_t Myfloat;
typedef enum {FALSE=0, TRUE=1} Bool;

// Constants
#define RADIX   10
#define BINARYPOINT 28

// SIGNIFWID is the "real" significand width, not including the
// phantom bit
#define SIGNIFWID 23
#define HALFWID 12
#define EXPWID 8
#define EXCESS 127

#define SIGNIFSHFT 0
#define EXPSHFT (SIGNIFSHFT + SIGNIFWID)
#define SIGNSHFT (EXPSHFT + EXPWID)
#define SIGNIFLOSHFT 0
#define SIGNIFHISHFT (SIGNIFLOSHFT + HALFWID)

#define PHANTOMMSK (1 << SIGNIFWID)
#define SIGNIFMSK (PHANTOMMSK - 1)
#define EXPMSK (((1 << EXPWID) - 1) << EXPSHFT)
#define SIGNMSK (1 << SIGNSHFT)
#define SIGNIFLOMSK ((1 << HALFWID) - 1)
#define SIGNIFHIMSK (SIGNIFLOMSK << SIGNIFHISHFT)
#define CARRYFROMMULT (1 << ((2*HALFWID)-1))
#define FRACMSK ((1 << BINARYPOINT) - 1)

// Handy macroes.

// The first two here are a bit flaky, since they only work if all the
// characters showing values are in order in the character
// set.  For character sets in the ISO-8859 family (including ASCII)
// that means the radix has to be 10 or less.
#define CHARTOVAL(c) ((c)-'0')
#define VALTOCHAR(v) ((v) + '0')

#define GETFIELD(val, mask, shift) (((val) & (mask)) >> (shift))
#define ORFIELD(oldval, fieldval, shift) ((oldval) | ((fieldval) << (shift)))

#define GETSIGN(val) GETFIELD(val, SIGNMSK, SIGNSHFT)
#define GETEXP(val) (GETFIELD(val, EXPMSK, EXPSHFT) - EXCESS)
#define GETSIGNIF(val) (GETFIELD(val, SIGNIFMSK, SIGNIFSHFT) | PHANTOMMSK)
#define GETSIGNIFLO(val) GETFIELD(GETSIGNIF(val), SIGNIFLOMSK, SIGNIFLOSHFT)
#define GETSIGNIFHI(val) GETFIELD(GETSIGNIF(val), SIGNIFHIMSK, SIGNIFHISHFT)

#define GETFRAC(val) GETFIELD(val, FRACMSK, 0)
#define GETINT(val) GETFIELD(val, ~FRACMSK, BINARYPOINT)

#define ASSEMBLE(sign, exponent, significand) (((sign) << SIGNSHFT) | \
                                               (((exponent) + EXCESS) << EXPSHFT) | \
                                               (((significand) & SIGNIFMSK) << SIGNIFSHFT))
// Forward declarations
Myfloat readfloat(void);
uint32_t readfrac(unsigned char inch);
Myfloat multiply(Myfloat op1, Myfloat op2);
void    writefloat(Myfloat val);
void    writeint(uint32_t intval);

/*
  Main function:  driver for the rest.

  No command-line arguments.

  Return value is 0.
*/

int main()
{
    Myfloat op1, op2, res;

    op1 = readfloat();
    op2 = readfloat();
    res = multiply(op1, op2);
#ifdef EBUG 
    printf("op1 %0x08x, op2 %0x08x, res %0x08x\n", op1, op2, res);
#endif
    writefloat(res);

    return 0;
}
    
/*
  readfloat():  read a floating point number from stdin

  Input format as given in assignment.
  
  Assumes int part and fraction part can be stored in a 32 bit integer,
  will not handle anything requiring a denorm (including 0.0).
  
  Performs no input checking; assumes data matches specification.
  There may be inputs that aren't as specified which this will also
  handle correctly; that would be a coincidence.
  
*/
Myfloat readfloat()
{
    Myfloat retval;
    
    unsigned char inch;
    Bool negative = FALSE;
    uint32_t signif = 0;
    int exponent = 0;

    uint32_t intpart = 0, fracpart = 0, significand;
    uint32_t tmpint, tmpfrac;
    
    //  int scale = RADIX;

    // get past any leading junk
    do {
        inch = getchar();
    } while ((inch != '-') && !isdigit(inch));

    // if first character is a minus sign, we're negative.
    if (inch == '-') {
        negative = TRUE;
        inch = getchar();
    }

    // now read the integer part.
    do {
        intpart = intpart * RADIX + CHARTOVAL(inch);
        inch = getchar();
    } while (isdigit(inch));


    // note that when we get here, inch is the decimal point.  Another
    // getchar() gets the first digit right of decimal point.
    fracpart = readfrac(getchar());
    
    // Here's my original fraction conversion code
    //  do {
    //      fracpart = fracpart + (CHARTOVAL(inch) << BINARYPOINT)/scale;
    //      scale *= RADIX;
    //      inch = getchar();
    //  } while (isdigit(inch));

    // Now that we're parsed, it's time to get to work by calculating
    // the exponent and significand.  We have two
    // cases, for a number greater than or equal to 1 vs. a number
    // less than 1.
    if (intpart != 0) {
        // first case:  not less than 1.

        // figure out what the exponent is
        tmpint = intpart;
        exponent = 0;
        while (tmpint != 1) {
            tmpint >>= 1;
            exponent++;
        }

        // construct significand
        significand =
          (intpart << (EXPSHFT - exponent)) |
          ((fracpart >> (BINARYPOINT - EXPSHFT + exponent)));
        
    } else {
        // second case:  less than 1.
        tmpfrac = fracpart;
        exponent = 0;
        while (GETINT(tmpfrac) == 0) {
            exponent--;
            tmpfrac <<= 1;
        }

        // now we have two more cases:  do we shift the fraction left
        // (exponent magnitude greater than 4), or right (exponent
        // magnitude less than 4).
        if (exponent < (EXPSHFT-BINARYPOINT))
            significand = fracpart << ((EXPSHFT-BINARYPOINT) - exponent);
        else
            significand = fracpart >> (exponent - (EXPSHFT-BINARYPOINT));
    }

    retval = ASSEMBLE(negative, exponent, significand);

    // and return
    return retval;
}

// When asked, I told a student they could use the algorithm for
// fraction conversion that I've got commented out above.  It wasn't
// until I was testing my code that I realized that the approach
// causes *nasty* rounding errors -- I'll stand by my answer and
// accept the algorithm above, but here's a recursive algorithm that
// works a whole lot better
uint32_t readfrac(unsigned char inch)
{
    uint32_t tmp;
    
    if (!isdigit(inch))
        return 0;
    else {
        tmp = readfrac(getchar());
        tmp = tmp + (CHARTOVAL(inch) << BINARYPOINT);
        tmp = tmp / RADIX;
        return tmp;
        //      return (readfrac(getchar()) + (CHARTOVAL(inch) << BINARYPOINT))/RADIX;
    }
}

/*
  multiply()

  Takes two floating point operands and multiplies them, producing
  result

  Assumes single precision (notably, doesn't assume promotion to
  double precision as in "real" C floating point

  No input checking (no infinite, NaN, denorm, etc)
  
  No result checking (overflow, underflow, etc)

  No rounding
*/
Myfloat multiply(Myfloat op1, Myfloat op2)
{
    Myfloat retval;
    Bool sign;
    int exponent;
    uint32_t significand;
    uint32_t op1hi, op1lo, op2hi, op2lo;

#ifdef EBUG
    uint32_t signif, t1, t2, t3, t4, r1, r2, r3;
#endif

    // compute sign and tentative exponent of result
    sign = GETSIGN(op1) ^ GETSIGN(op2);
    exponent = GETEXP(op1) + GETEXP(op2);

#ifdef EBUG 
    signif = GETSIGNIF(op1);
    signif = GETSIGNIF(op2);
    
    t1 = GETSIGNIFLO(op1);
    t2 = GETSIGNIFLO(op2);
    t3 = GETSIGNIFHI(op1);
    t4 = GETSIGNIFHI(op2);

    r1 = t1 * t2;
    r2 = (r1 >> 12) + t1*t3 + t2*t4;
    r3 = (r2 >> 12) + t3*t4;
#endif
    
    // significand
    significand =
      (((GETSIGNIFLO(op1)*GETSIGNIFLO(op2)) >> HALFWID) +
       ((GETSIGNIFLO(op1)*GETSIGNIFHI(op2) + GETSIGNIFHI(op1)*GETSIGNIFLO(op2))) >> HALFWID) +
      GETSIGNIFHI(op1)*GETSIGNIFHI(op2);

    // renormalize.
    if (significand & CARRYFROMMULT)
        exponent++;
    else
        significand <<= 1;

    retval = ASSEMBLE(sign, exponent, significand);
    
    return retval;
}

/*
  writefloat()

  Write a floating point value to stdout.

  Assumes single precision (notably, doesn't assume promotion to
  double precision as in "real" C floating point

  Assumes integer and fraction parts each fit in a word
  
  No input checking (no infinite, NaN, denorm, etc)
*/
void writefloat(Myfloat val)
{
    int exponent;
    uint32_t significand;
    uint32_t intpart, fracpart;

    exponent = GETEXP(val);
    significand = GETSIGNIF(val);
    
    // print sign if negative
    if (GETSIGN(val) == 1)
        putchar('-');

    // if the exponent is less than 0, the integer part will be 0.
    // We'll just print that.
    if (exponent < 0)
        putchar('0');
    else {
        // special cases time...  if the exponent is less than the
        // siginificand width, we have to right-shift.  If it's
        // greater, we have to left-shift (I could swear that at one
        // time C allowed shifts by negative amounts.  Last time I
        // tried it, the compiler got unhappy at me).
        if (exponent < SIGNIFWID)
            intpart = significand >> (SIGNIFWID - exponent);
        else
            intpart = significand << (exponent - SIGNIFWID);

        writeint(intpart);
    }
        
    putchar('.');

    // More special cases (you know, putting the binary point for
    // conversions at the same place as in the fp representation would
    // save a lot of work.  But since this is what I suggested to you,
    // I'll do it too as some sort of misguided penance).  We may need
    // to either left-shift or right-shift...
    if (exponent > SIGNIFWID-BINARYPOINT)
        fracpart = significand << (exponent - (SIGNIFWID-BINARYPOINT));
    else
        fracpart = significand >> ((SIGNIFWID-BINARYPOINT) - exponent);

    fracpart = GETFRAC(fracpart);
    
    // print 0 if fraction part is 0 or print fraction part
    if (fracpart == 0)
        putchar('0');
    else
        while (fracpart != 0) {
            fracpart = fracpart * RADIX;
            putchar(VALTOCHAR(GETINT(fracpart)));
            fracpart = GETFRAC(fracpart);
        }
    
    // print newline
    putchar('\n');

    return;
}

void writeint(uint32_t intval)
{
    if (intval == 0)
        return;

    writeint(intval / RADIX);
    putchar(VALTOCHAR(intval % RADIX) );

    return;
}
