/*****************************************************************************/
/*                                                                           */
/*    CS 473 -- Fall, 2004                                                   */
/*    HW2 Sample Solution                                                    */
/*                                                                           */
/*    This program will decode and print the contents of MIPS instructions   */
/*    Input and output are as specified in the assignment                    */
/*                                                                           */
/*    This isn't quite going to match the spec for HW1, since I want the     */
/*    code to be as useful as possible for people needing help starting      */
/*    HW2.                                                                   */
/*                                                                           */
/*****************************************************************************/
#include <assert.h>
#include <stdint.h>
#include <stdio.h>

typedef uint32_t Instruction;
typedef uint32_t Address;
typedef int16_t AddImm;
typedef uint8_t Regnum;       // should really be five bits
typedef uint8_t Opcode;       // should really be six bits
typedef uint8_t Funct;        // should really be six bits
typedef uint8_t Jump;		  // should really be 26 bits
typedef enum {FALSE=0, TRUE=1} Bool;

// size of instruction memory
#define INSTMEMSIZE 1024

// sizes of fields in instruction
#define OPCODESIZE 6
#define REGNUMSIZE 5
#define ADDIMMSIZE 16
#define FUNCSIZE   6
#define SHIFTSIZE  5
#define JUMPSIZE   26

// handy mask/shift macro
#define EXTRACT(input, width, shift) (((input) >> (shift)) & ((1 << width)-1))

// macroes to extract fields from instruction (defined right to left)
#define FUNC(inst)      EXTRACT((inst), (FUNCSIZE), 0)
#define ADDIMM(inst)	EXTRACT((inst), (ADDIMMSIZE), 0)
#define ADDIMMSIGN(inst) EXTRACT((inst), 1, ((ADDIMMSIZE)-1))
#define JUMP(inst)		EXTRACT((inst), (JUMPSIZE), 0)

#define RD(inst)        EXTRACT((inst), (REGNUMSIZE), (FUNCSIZE+SHIFTSIZE))
#define RT(inst)        EXTRACT((inst), (REGNUMSIZE), (ADDIMMSIZE))
#define RS(inst)        EXTRACT((inst), (REGNUMSIZE), (REGNUMSIZE+ADDIMMSIZE))
#define OPCODE(inst)    EXTRACT((inst), (OPCODESIZE), (JUMPSIZE))

//  macroes for byte-to-word conversions
#define BYTETOWORD(byteaddr) ((byteaddr) >> 2)
#define WORDTOBYTE(wordaddr) ((wordaddr) << 2)

// sign and zero extension
#define SIGNEXTEND(val) ((int32_t) (ADDIMMSIGN(val) ? ((val) | 0xffff0000) : ((val) & 0x0000ffff)))
#define ZEROEXTEND(val) ((uint32_t)((val) & 0x0000ffff))

// the opcodes we're dealing with
#define RFORMAT	0
#define ADDI    8
#define ADDIU   9
#define	LW		35
#define	BEQ		4
#define	J		2

// function fields for R format instructions
#define	ADD		32
#define	SUB		34

// Forward declarations.  Notice that instructionMemory is being
// defined in a way that will work well for dataMemoory as well
// (though with more duplicated code than I really like...)
void instructionMemory(Address addr,
					   Bool memRead, Bool memWrite,
					   Instruction *data);

void decodeInstruction(Address instrAddr, Instruction IR);

// Main program.  Just two loops:  one reads stdin to get all the
// instructions, the other goes through all of instruction memory
// decoding them.
int main()
{
	Instruction InstructionRegister;
	Address PC;

	// read in all the instructions....
	while (scanf("%x %x", &PC, &InstructionRegister) != EOF)
		instructionMemory(PC, FALSE, TRUE, &InstructionRegister);
	
	for (PC = 0; PC < INSTMEMSIZE; PC += 4) {
		instructionMemory(PC, TRUE, FALSE, &InstructionRegister);
		decodeInstruction(PC, InstructionRegister);
	}

	return 0;
}

// instruction memory capable of both reading and writing (note that
// the instruction memory in the figures can only read, and has no
// control input.  Apparently instructions land in it through
// telepathy).  This one is basically the same code as for the data
// memory.
void instructionMemory(Address addr,
					   Bool memRead, Bool memWrite,
					   Instruction *data)
{
	static Instruction memory[INSTMEMSIZE] = {0};;

	assert(BYTETOWORD(addr) < INSTMEMSIZE);
	assert((memRead == FALSE) || (memWrite == FALSE));

	if (memRead)
		*data = memory[BYTETOWORD(addr)];

	if (memWrite)
		memory[BYTETOWORD(addr)] = *data;

	return;
}

// decode and print an instruction.  This is table-driven as much as
// possible;  the switch statements determine the instruction format,
// and then we use table lookups to print the instruction names.
// Takes more typing, but much easier to debug.
void decodeInstruction(Address addr, Instruction instr)
{
	// array of translations from opcode to instruction name.  With
	// this few instructions we'd be better off with a little
	// function, but this is more extensible...  too bad C doesn't
	// have associative arrays!
	static char *opcodeToName[1 << OPCODESIZE] = {
	  NULL,        //  0 -- if this turns up, something is wrong
	  NULL,        //  1
	  "j",         //  2
	  NULL,        //  3
	  "beq",       //  4
	  NULL,        //  5
	  NULL,        //  6
	  NULL,        //  7
	  "addi",      //  8
	  "addiu",     //  9
	  NULL,        // 10
	  NULL,        // 11
	  NULL,        // 12
	  NULL,        // 13
	  NULL,        // 14
	  NULL,        // 15
	  NULL,        // 16
	  NULL,        // 17
	  NULL,        // 18
	  NULL,        // 19
	  NULL,        // 20
	  NULL,        // 21
	  NULL,        // 22
	  NULL,        // 23
	  NULL,        // 24
	  NULL,        // 25
	  NULL,        // 26
	  NULL,        // 27
	  NULL,        // 28
	  NULL,        // 29
	  NULL,        // 30
	  NULL,        // 31
	  NULL,        // 32
	  NULL,        // 33
	  NULL,        // 34
	  "lw"         // 35
	};
	
	// similar array for func fields
	static char *funcToName[1 << FUNCSIZE] = {
	  NULL,        //  0 -- if this turns up, something is wrong
	  NULL,        //  1
	  NULL,        //  2
	  NULL,        //  3
	  NULL,        //  4
	  NULL,        //  5
	  NULL,        //  6
	  NULL,        //  7
	  NULL,        //  8
	  NULL,        //  9
	  NULL,        // 10
	  NULL,        // 11
	  NULL,        // 12
	  NULL,        // 13
	  NULL,        // 14
	  NULL,        // 15
	  NULL,        // 16
	  NULL,        // 17
	  NULL,        // 18
	  NULL,        // 19
	  NULL,        // 20
	  NULL,        // 21
	  NULL,        // 22
	  NULL,        // 23
	  NULL,        // 24
	  NULL,        // 25
	  NULL,        // 26
	  NULL,        // 27
	  NULL,        // 28
	  NULL,        // 29
	  NULL,        // 30
	  NULL,        // 31
	  "add",       // 32
	  NULL,        // 33
	  "sub"        // 34
	};

	// print address and raw instruction if this is a defined isntruction
	if ((opcodeToName[OPCODE(instr)] != NULL) ||
		((OPCODE(instr) == RFORMAT) &&
		 funcToName[FUNC(instr)] != NULL))
		printf("%08x %08x ", addr, instr);
	
	switch (OPCODE(instr)) {
		// R-format instructions
	  case RFORMAT:
		switch FUNC(instr) {
		  case ADD:
		  case SUB:
			printf("%s $%d,$%d,$%d\n", funcToName[FUNC(instr)], RD(instr), RS(instr), RT(instr));
			break;

		  default:
			// unhandled instruction
			break;
		}
		break;

		// I-format arithmetic instructions, no sign-extend
	  case ADDIU:
		printf("%s $%d,$%d,%d\n", opcodeToName[OPCODE(instr)], RT(instr), RS(instr), ZEROEXTEND(ADDIMM(instr)));
		break;

		// I-format arithmetic instructions, with sign-extend
		// oops!  I didn't ask you do do any of these!  Oh well, I
		// will anyway
	  case ADDI:
		printf("%s $%d,$%d,%d\n", opcodeToName[OPCODE(instr)], RT(instr), RS(instr), SIGNEXTEND(ADDIMM(instr)));
		break;
		
		// I-format memory instructions
	  case LW:
		printf("%s $%d,%d($%d)\n", opcodeToName[OPCODE(instr)], RT(instr), SIGNEXTEND(ADDIMM(instr)), RS(instr));
		break;

		// I-format branch instructions
	  case BEQ:
		printf("%s $%d,$%d,%d\n", opcodeToName[OPCODE(instr)], RS(instr), RT(instr), WORDTOBYTE(SIGNEXTEND(ADDIMM(instr))));
		break;

		// J-format instructions
	  case J:
		printf("%s %d\n", opcodeToName[OPCODE(instr)], WORDTOBYTE(JUMP(instr)));
		break;

	  default:
		// unhandled instruction
		break;
	}

	return;
}
