// Simulation of simple MIPS implementation
// Sample solution to CS 473 HW 4
// Start with HW 3, add pipeline registers
// Urrgh, I didn't realize Figure 5.24 was a better starting point for
// you!  Though it's got typoes...
// I'm implementing PC using the technique I'm recommending you use
// for the pipeline registers.
#include <assert.h>
#include <stdint.h>
#include <stdio.h>

// constants
#define INSTMEMSIZE 1024
#define DATAMEMSIZE 1024
#define OPCODEWID 6
#define REGWID 5
#define SHIFTWID 5
#define FUNCWID 6
#define ADDRIMMWID 16
#define JADDRWID 26
#define PCMASK 0xf0000000

#define NUMREGS (1 << REGWID)
#define NUMELEMS(array) (sizeof(array)/sizeof(array[0]))
#define MAXMUX 3

#define SIGN16 0x00008000
#define HIGH16 0xffff0000
#define LOWS16  0x0000ffff

// instruction disassembly
#define MASK(in, size, shift) (((in) >> (shift)) & ((1 << (size)) - 1))
#define FUNC(instr) MASK(instr, FUNCWID, 0)
#define ADDRIMM(instr) MASK(instr, ADDRIMMWID, 0)
#define JADDR(instr) MASK(instr, JADDRWID, 0)
#define RD(instr) MASK(instr, REGWID, FUNCWID+SHIFTWID)
#define RT(instr) MASK(instr, REGWID, FUNCWID+SHIFTWID+REGWID)
#define RS(instr) MASK(instr, REGWID, FUNCWID+SHIFTWID+2*REGWID)
#define OPCODE(instr) MASK(instr, OPCODEWID, FUNCWID+SHIFTWID+3*REGWID)
#define PCHIGHORDER(pc) ((pc) & PCMASK)

// reporting.  We'll leave the old defs in here for debugging
#ifdef EBUG
#define UNIT(name) printf("\n    " #name "\n")
#define INPUTS() printf("        Inputs:\n");
#define REPORTIN(name,  digits)     printf("            %s 0x%0" #digits "x\n", #name, name)
#define OUTPUTS() printf("        Outputs:\n");
#define REPORTOUT(name, digits)     if ((name) != NULL) printf("            %s 0x%0" #digits "x\n", #name, *name)
#else
#define UNIT(name)
#define INPUTS()
#define REPORTIN(name,  digits)
#define OUTPUTS()
#define REPORTOUT(name, digits)
#endif

// helper for memory accesses
#define BYTETOWORD(val) ((val) >> 2)

// types
typedef uint32_t Address;
typedef uint32_t Instruction;
typedef uint32_t Data;
typedef unsigned int OpCode; // really should be six bits
typedef unsigned int RegNo;  // really should be five bits
typedef unsigned int MuxSel; // width really depends on mux type
typedef unsigned int Func;   // really should be six bits

// easier to read control lines when they're enums where appropriate.
typedef enum {FALSE=0, TRUE=1} Bool;
typedef enum {BRNONE=0, BRBEQ=1} BranchType;
typedef enum {JNO=0, JYES=1} JMuxSel;
typedef enum {READ=0, WRITE=1} InstrMemOp;
typedef enum {CTRLADD=0, CTRLSUB=1, CTRLFUNC=2} AluOp;
typedef enum {ALUAND=0, ALUOR=1, ALUADD=2, ALUSUB=6, ALUSLT=7, ALUNOR=0xc} AluCtrl;
typedef enum {FROMRT=0, FROMRD=1} RegDstSel;
typedef enum {FROMREG=0, FROMSE=1, FROMCONST=2} AluSrcSel;
typedef enum {FROMALU=0, FROMMEM=1} MemToRegSel;
typedef enum {FROMADDRINC=0, FROMBRANCHADD=1, FROMJADD=2} BranchMuxSel;

// reporting
void report(uint32_t pc, uint32_t registers[NUMREGS], uint32_t data_memory[DATAMEMSIZE]);

// functional unit forward declarations.

void memOp(Data memory[], unsigned int memsize,
      Address address, Data writeData,
      Bool memRead, Bool memWrite,
      Data *readData);
void alu(uint32_t op1, uint32_t op2, AluCtrl operation, Bool *zero, uint32_t *result);
void control(OpCode opcode, RegDstSel *regDst, JMuxSel *jump, BranchType *branch, Bool *memRead,
             MemToRegSel *memToReg, AluOp *aluOp, Bool *memWrite, AluSrcSel *aluSrc, Bool *regWrite);
void registers(Data regs[NUMREGS], RegNo readReg1, RegNo readReg2, RegNo writeReg, Bool regWrite, Data writeData, Data *readData1, Data *readData2);
void signExtend(Data in, Data *out);
void shiftLeft2(Data in, Data *out);
void jCombine(Address fromPC, Address fromJfield, Address *jumpAddr);
void mux(unsigned int inputs, Data in0, Data in1, Data in2, MuxSel whichone, Data *out);
void aluControl(AluOp aluOp, Func func, AluCtrl *aluControlOut);
void and(Bool in1, Bool in2, Bool *out);
void pc(Address in, Address *out);

// pipeline units
void ifId(Instruction instrMemOut, Instruction *ifId_instrMemOut,
		  Address  pcAluOut,	   Address     *ifId_pcAluOut);
void idEx(RegDstSel regDst,        RegDstSel *idEx_regDst,
		  AluOp     aluOp,		   AluOp     *idEx_aluOp,
		  AluSrcSel aluSrc,		   AluSrcSel *idEx_aluSrc,
		  Bool      branch,		   Bool      *idEx_branch,
		  Bool      memRead,	   Bool      *idEx_memRead,
		  Bool      memWrite,	   Bool      *idEx_memWrite,
		  Bool      memToReg,	   Bool      *idEx_memToReg,
		  Bool      regWrite,	   Bool      *idEx_regWrite,
		  Bool      jump,		   Bool      *idEx_jump,

		  Address   ifId_pcAluOut, Address   *idEx_pcAluOut,

		  Data      readData1,     Data      *idEx_readData1,
		  Data      readData2,     Data      *idEx_readData2,
		  Data      signExtendOut, Data      *idEx_signExtendOut,
		  Data      jAddress,      Data      *idEx_jAddress,
		  Data      zeroExtendOut, Data      *idEx_zeroExtendOut,
		  RegNo     rdField,       RegNo    *idEx_rdField,
		  RegNo     rtField,       RegNo    *idEx_rtField);

void exMem(Bool    idEx_branch,    Bool    *exMem_branch,
		   Bool    idEx_jump,      Bool    *exMem_jump,
		   Bool    idEx_memRead,   Bool    *exMem_memRead,
		   Bool    idEx_memWrite,  Bool    *exMem_memWrite,
		   Bool    idEx_memToReg,  Bool    *exMem_memToReg,
		   Bool    idEx_regWrite,  Bool    *exMem_regWrite,

		   Address jCombineOut,       Address *exMem_jCombineOut,
		   Address branchAddrOut,  Address *exMem_branchAddrOut,

		   Bool   zero,                 Bool    *exMem_zero,
		   Data   aluResult,            Data    *exMem_aluResult,
		   Data   idEx_readData2,       Data    *exMem_readData2,
		   RegNo  writeRegMuxOut,       RegNo  *exMem_writeRegMuxOut);
void memWb(Bool   exMem_memToReg,       Bool   *memWb_memToReg,
		   Bool   exMem_regWrite,       Bool   *memWb_regWrite,
		   RegNo  exMem_writeRegMuxOut, RegNo *memWb_writeRegMuxOut,

		   Data   readData,             Data   *memWb_readData,
		   Data   exMem_aluResult,      Data   *memWb_aluResult);



// other forward declarations
void readProgram(Instruction instructionMemory[INSTMEMSIZE]);

int main()
{
    // all the wires (sorted by pipeline stage).  Within each stage,
    // control wires will be first, then wires related to branch/jump,
    // then "ordinary" data wires.  Wires coming out of pipeline
    // registers will have names starting with name of pipeline reg.

	// Instruction fetch
    Address pcOut = 0;                  // from PC to instruction memory
    Instruction instrMemOut = 0;        // output from instruction memory
    Address pcAluOut = 0;               // from alu incrementer to jump combiner and branch mux

	// Decode/register read
    RegDstSel regDst = 0;               // is result register selected by
									// rt or rd?
    AluOp aluOp = 0;                    // ALU operation
    AluSrcSel aluSrc = 0;               // second alu input from registers,
                                    // sign-extended imm, or non-sign-extended imm?
    Bool branch = 0;                    // is this a branch instruction?
    Bool memRead = 0;                   // read data memory?
    Bool memWrite = 0;                  // write data memory?
    MemToRegSel memToReg = 0;           // register output from alu or from data memory?
    Bool regWrite = 0;                  // write result to registers?
    JMuxSel jump = 0;                   // take a jump?

	Address ifId_pcAluOut = 0;
	Instruction ifId_instrMemOut = 0;

    Data readData1 = 0, readData2 = 0;      // from registers to alu (or alu input mux)
    Data signExtendOut = 0;             // from sign extender to alu input mux

	// Execute
    RegDstSel idEx_regDst = 0;
    AluOp idEx_aluOp = 0;
    AluSrcSel idEx_aluSrc = 0;
    Bool idEx_branch = 0;
    JMuxSel idEx_jump = 0;
    Bool idEx_memRead = 0;
    Bool idEx_memWrite = 0;
    MemToRegSel idEx_memToReg = 0;
    Bool idEx_regWrite = 0;

    AluCtrl aluControlOut = 0;          // output from alu control unit;
                                    // contains acutal alu command

	Address idEx_pcAluOut = 0;
	Data idEx_readData1 = 0, idEx_readData2 = 0;
	Data idEx_signExtendOut = 0;
	Data idEx_jAddress = 0;
	Data idEx_zeroExtendOut = 0;
	RegNo idEx_rdField = 0, idEx_rtField = 0;
    RegNo writeRegMuxOut = 0;          // from write register number mux
                                    // to write register number input
                                    // on registers

	Address jAddrOut = 0;               // output from jump left-shifter
    Address jCombineOut = 0;            // combine old PC with
									// left-shifted jump addr
    Address branchShifterOut = 0;       // output of branch shifter
    Data aluMuxOut = 0;                 // second input to ALU
    Address branchAddrOut = 0;          // branch target address

    Bool zero = 0;                      // ALU result 0?
    Data aluResult = 0;                 // ALU result

	// Memory
    Bool exMem_branch = 0;                    // is this a branch instruction?
    Bool exMem_memRead = 0;                   // read data memory?
    Bool exMem_memWrite = 0;                  // write data memory?
    MemToRegSel exMem_memToReg = 0;           // register output from alu or from data memory?
    Bool exMem_regWrite = 0;                  // write result to registers?
    BranchMuxSel branchMuxControl = 0;        // take a branch?
    JMuxSel exMem_jump = 0;                   // take a jump?

	Address exMem_jCombineOut = 0;
	Address exMem_branchAddrOut = 0;

    Address jMuxOut = 0;                // from jump mux to PC
    Address branchMuxOut = 0;           // from branch mux to jump mux

	RegNo exMem_writeRegMuxOut = 0;

	Bool exMem_zero = 0;
	Data exMem_aluResult = 0;
	Data exMem_readData2 = 0;
    Data readData = 0;                  // from data mem to alu/mem
	                                // mux

	// Writeback
	Bool memWb_memToReg = 0;
	Bool memWb_regWrite = 0;

	RegNo memWb_writeRegMuxOut = 0;

	Data memWb_aluResult = 0;
	Data memWb_readData = 0;

    Data aluMemMuxOut = 0;              // from alu/mem mux to register
                                    // write data input

	// registers and memory
	Data dataMemory[DATAMEMSIZE] = {0};
	Data instructionMemory[INSTMEMSIZE]={0};
	Data regs[NUMREGS] = {0};

    // now the code.

    // read in the program...
    readProgram(instructionMemory);

    // interestingly, MIPS has no HALT instruction.  So this program
    // should never terminate.  Funny nobody asked me about this....
    // I'm going to put in a hard-loop detector:  if the PC is ever
    // given the same address as the last instruciton executed, I'll
    // end the program.
    while (1) {
		report(pcOut, regs, dataMemory);

        // the sequence of these operations must be subject to the
        // constraint that all data is available before it is used.
        // I'm trying to actually call them in the order you'd read
        // them off the figure, going by columns.  This will make
        // pipelining easier later.

        // instruction fetch stage
        UNIT(PC Incrementer (alu));
        alu(pcOut, 4, ALUADD, NULL, &pcAluOut);

        UNIT(Instruction Memory);
		memOp(instructionMemory, INSTMEMSIZE, pcOut, 0, TRUE, FALSE, &instrMemOut);
		// detect end-of-program
		if (instrMemOut == 0)
			return 0;

        // decode and read registers stage
        UNIT(Control Unit);
        control(OPCODE(ifId_instrMemOut),
                &regDst, &jump, &branch, &memRead, &memToReg, &aluOp, &memWrite,
                &aluSrc, &regWrite);

		// registers will be handled in WB stage (needed for some data dependencies

        UNIT(Sign Extender);
        signExtend(ADDRIMM(ifId_instrMemOut), &signExtendOut);

        // alu and branch calc stage
        UNIT(Branch left-shifter);
        shiftLeft2(idEx_signExtendOut, &branchShifterOut);

        UNIT(Jump left-shifter);
        shiftLeft2(JADDR(ifId_instrMemOut), &jAddrOut);

        UNIT(Jump address combiner);
        jCombine(jAddrOut, PCHIGHORDER(pcOut), &jCombineOut);

        UNIT(ALU Source Mux);
        mux(3, idEx_readData2, idEx_signExtendOut, idEx_zeroExtendOut, idEx_aluSrc, &aluMuxOut);

        UNIT(ALU Control);
        aluControl(idEx_aluOp, FUNC(idEx_signExtendOut), &aluControlOut);

        UNIT(Branch Target Adder);
        alu(idEx_pcAluOut, branchShifterOut, ALUADD, NULL, &branchAddrOut);

        UNIT(Main ALU);
        alu(idEx_readData1, aluMuxOut, aluControlOut, &zero, &aluResult);

        UNIT(Write Register Mux);
        mux(2, idEx_rtField, idEx_rdField, 0, idEx_regDst, &writeRegMuxOut);

        // branch decision and mem op stage
        UNIT(BEQ And);
        and(exMem_branch, exMem_zero, &branchMuxControl);

        UNIT(Branch Mux);
        mux(2, pcAluOut, exMem_branchAddrOut, 0, branchMuxControl, &branchMuxOut);

        UNIT(Jump mux);
        mux(2, branchMuxOut, exMem_jCombineOut, 0, exMem_jump, &jMuxOut);

        UNIT(Data Memory);
		memOp(dataMemory, DATAMEMSIZE, exMem_aluResult, exMem_readData2, exMem_memRead, exMem_memWrite, &readData);

        // writeback stage
        UNIT(ALU/Memory Mux);
        mux(2, memWb_aluResult, memWb_readData, 0, memWb_memToReg, &aluMemMuxOut);

        UNIT(Registers);
        registers(regs, RS(ifId_instrMemOut), RT(ifId_instrMemOut), memWb_writeRegMuxOut, memWb_regWrite, aluMemMuxOut,
                  &readData1, &readData2);

        UNIT(Program Counter);
		pc(jMuxOut, &pcOut);

		// pipeline registers.  I have to go right to left here to
		// keep from propagating values too far when there's just a
		// wire in a stage (I guess I should really make the wires be
		// units....)
		UNIT(memWb);
		memWb(exMem_memToReg,       &memWb_memToReg,
			  exMem_regWrite,       &memWb_regWrite,
			  exMem_writeRegMuxOut, &memWb_writeRegMuxOut,

			  readData,             &memWb_readData,
			  exMem_aluResult,      &memWb_aluResult);
		
		UNIT(exMem);
		exMem(idEx_branch,    &exMem_branch,
			  idEx_jump,      &exMem_jump,
			  idEx_memRead,   &exMem_memRead,
			  idEx_memWrite,  &exMem_memWrite,
			  idEx_memToReg,  &exMem_memToReg,
			  idEx_regWrite,  &exMem_regWrite,

			  jCombineOut,       &exMem_jCombineOut,
			  branchAddrOut,  &exMem_branchAddrOut,

			  zero,           &exMem_zero,
			  aluResult,      &exMem_aluResult,
			  idEx_readData2, &exMem_readData2,
			  writeRegMuxOut, &exMem_writeRegMuxOut);

		UNIT(idEx);
		idEx(regDst,                       &idEx_regDst,
			 aluOp,						   &idEx_aluOp,
			 aluSrc,					   &idEx_aluSrc,
			 branch,					   &idEx_branch,
			 memRead,					   &idEx_memRead,
			 memWrite,					   &idEx_memWrite,
			 memToReg,					   &idEx_memToReg,
			 regWrite,					   &idEx_regWrite,
			 jump,						   &idEx_jump,

			 ifId_pcAluOut,				   &idEx_pcAluOut,

			 readData1,                    &idEx_readData1,
			 readData2,                    &idEx_readData2,
			 signExtendOut,                &idEx_signExtendOut,
			 JADDR(ifId_instrMemOut),      &idEx_jAddress,
			 ADDRIMM(ifId_instrMemOut),    &idEx_zeroExtendOut,
			 RD(ifId_instrMemOut),         &idEx_rdField,
			 RT(ifId_instrMemOut),         &idEx_rtField);

		UNIT(ifId);
		ifId(instrMemOut, &ifId_instrMemOut,
			 pcAluOut,	  &ifId_pcAluOut);

    }
}

/*****************************************************************************/
/*                                                                           */
/*  readProgram                                                              */
/*                                                                           */
/*      reads program from standard input.                                   */
/*                                                                           */
/*      no parameters                                                        */
/*                                                                           */
/*****************************************************************************/
void readProgram(Instruction instructionMemory[INSTMEMSIZE])
{
    Address PC;
    Instruction InstructionRegister;

    // read in all the instructions....
    while (scanf("%x %x", &PC, &InstructionRegister) != EOF)
        memOp(instructionMemory, INSTMEMSIZE, PC, InstructionRegister, FALSE, TRUE, NULL);

}

/*****************************************************************************/
/*                                                                           */
/*  alu                                                                      */
/*                                                                           */
/*      simulates an ALU.                                                    */
/*                                                                           */
/*      inputs:                                                              */
/*          op1, op2:  alu inputs                                            */
/*          operation: operation to be performed on the inputs               */
/*                                                                           */
/*      outputs:                                                             */
/*          zero:  boolean indicating whether result is 0                    */
/*          result: result of alu operation                                  */
/*                                                                           */
/*      I'm using the same code for all three places that call for an ALU.   */
/*      In two of them, it is just an adder, and I don't have to worry       */
/*      the zero output, but it keeps me from having to write extra code.    */
/*      Because of that, I'm allowing a null pointer on the zero output;     */
/*      rather than aborting on a null pointer there I just don't write to   */
/*      it                                                                   */
/*                                                                           */
/*****************************************************************************/
void alu(uint32_t op1, uint32_t op2, AluCtrl operation, Bool *zero, uint32_t *result)
{
    INPUTS();
    REPORTIN(op1, 8);
    REPORTIN(op2, 8);
    REPORTIN(operation, 1);

    assert(result != NULL);

    switch (operation) {
      case ALUAND:
        *result = op1 & op2;
        break;

      case ALUOR:
        *result = op1 | op2;
        break;

      case ALUADD:
        *result = op1 + op2;
        break;

      case ALUSUB:
        *result = op1 - op2;
        break;

      case ALUSLT:
        *result = op1 < op2;
        break;

      case ALUNOR:
        *result = ~(op1 | op2);
        break;

      default:
        assert(FALSE);
    }

    if (zero != NULL)
        *zero = (*result == 0);

    OUTPUTS();
    REPORTOUT(zero, 1);
    REPORTOUT(result, 8);
}

/*****************************************************************************/
/*                                                                           */
/*  memOp                                                                    */
/*                                                                           */
/*      common code shared between instructionMemory and dataMemory          */
/*      (in C++, this would be a base class...)                              */
/*                                                                           */
/*      inputs:                                                              */
/*          memory:  memory array being  simulated                           */
/*          memsize: size of memory array                                    */
/*          address: address to be read or written                           */
/*          writeData: data to be written                                    */
/*          memRead:  whether or not to read                                 */
/*          memWrite:  whether or not to write                               */
/*                                                                           */
/*      outputs:                                                             */
/*          readData:  data read from memory array                           */
/*                                                                           */
/*      Note that I do check for things like making sure you don't try tod   */
/*      read and write on the same cycle; also, I only check the readData    */
/*      pointer if you are actually going to read.                           */
/*                                                                           */
/*****************************************************************************/
void memOp(Data memory[], unsigned int memsize, Address address,
           Data writeData, Bool memRead, Bool memWrite, Data *readData)
{

    INPUTS();
    REPORTIN(address, 8);
    REPORTIN(writeData, 8);
    REPORTIN(memRead, 1);
    REPORTIN(memWrite, 1);

    assert((!memRead && !memWrite) || (BYTETOWORD(address) < memsize));
    assert(!memRead || (readData != NULL));
    assert(!memRead || !memWrite);

    if (memRead)
        *readData = memory[BYTETOWORD(address)];

    if (memWrite)
        memory[BYTETOWORD(address)] = writeData;

    OUTPUTS();
    REPORTOUT(readData, 8);
}

/*****************************************************************************/
/*                                                                           */
/*  control                                                                  */
/*                                                                           */
/*      main control unit                                                    */
/*                                                                           */
/*      Input:                                                               */
/*          opcode:  instruction op code                                     */
/*                                                                           */
/*      Outputs:                                                             */
/*          lots of control lines, given the same names as in the text       */
/*                                                                           */
/*      This control unit is actually simpler than the one from hw2; just    */
/*      a great big lookup table.  This assignment didn't actually call      */
/*      for a table that would let us implement the entire instruction set,  */
/*      but it seemed like the simplest way to do it.                        */
/*                                                                           */
/*****************************************************************************/
void control(OpCode opcode, RegDstSel *regDst, JMuxSel *jump, BranchType *branch, Bool *memRead,
        MemToRegSel *memToReg, AluOp *aluOp, Bool *memWrite, AluSrcSel *aluSrc, Bool *regWrite)
{
    const static struct {
        RegDstSel regDst;
        JMuxSel jump;
        BranchType branch;
        Bool memRead;
        MemToRegSel memToReg;
        AluOp aluOp;
        Bool memWrite;
        AluSrcSel aluSrc;
        Bool regWrite;
    } decode[] = {
    // regDst  jump  branch  memRead memToReg aluOp     memWrite aluSrc   regWrite
      {FROMRD, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, TRUE}, //  0 R-Format
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, //  1
      {FROMRT, JYES, BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, //  2 j
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, //  3
      {FROMRT, JNO,  BRBEQ,  FALSE,  FROMALU, CTRLSUB,  FALSE,   FROMREG, FALSE}, //  4 beq
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, //  5
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, //  6
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, //  7
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLADD,  FALSE,   FROMSE, TRUE}, //  8 addi
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLADD,  FALSE,   FROMCONST, TRUE}, //  9 addiu
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 10
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 11
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 12
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 13
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 14
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 15
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 16
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 17
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 18
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 19
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 20
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 21
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 22
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 23
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 24
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 25
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 26
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 27
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 28
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 29
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 30
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 31
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 32
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 33
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 34
      {FROMRT, JNO,  BRNONE, TRUE,   FROMMEM, CTRLADD,  FALSE,   FROMSE,  TRUE},  // 35 lw
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 36
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 37
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 38
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 39
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 40
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 41
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 42
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLADD,  TRUE,    FROMSE,  FALSE}, // 43 sw
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 44
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 45
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 46
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 47
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 48
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 49
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 50
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 51
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 52
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 53
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 54
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 55
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 56
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 57
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 58
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 59
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 60
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 61
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}, // 62
      {FROMRT, JNO,  BRNONE, FALSE,  FROMALU, CTRLFUNC, FALSE,   FROMREG, FALSE}  // 63
    };

    INPUTS();
    REPORTIN(opcode, 2);

    assert(opcode < NUMELEMS(decode));

    assert(regDst != NULL);
    assert(jump != NULL);
    assert(branch != NULL);
    assert(memRead != NULL);
    assert(memToReg != NULL);
    assert(aluOp != NULL);
    assert(memWrite != NULL);
    assert(aluSrc != NULL);
    assert(regWrite != NULL);

    *regDst   = decode[opcode].regDst;
    *jump     = decode[opcode].jump;
    *branch   = decode[opcode].branch;
    *memRead  = decode[opcode].memRead;
    *memToReg = decode[opcode].memToReg;
    *aluOp    = decode[opcode].aluOp;
    *memWrite = decode[opcode].memWrite;
    *aluSrc   = decode[opcode].aluSrc;
    *regWrite = decode[opcode].regWrite;

    OUTPUTS();
    REPORTOUT(regDst, 1);
    REPORTOUT(jump, 1);
    REPORTOUT(branch, 1);
    REPORTOUT(memRead, 1);
    REPORTOUT(memToReg, 1);
    REPORTOUT(aluOp, 1);
    REPORTOUT(memWrite, 1);
    REPORTOUT(aluSrc, 1);
    REPORTOUT(regWrite, 1);

}

/*****************************************************************************/
/*                                                                           */
/*  registers                                                                */
/*                                                                           */
/*      register set                                                         */
/*                                                                           */
/*      inputs:                                                              */
/*          regs:      register file                                         */
/*          readReg1, readReg2:  registers to read                           */
/*          writeReg:  register to write (if any)                            */
/*          regWrite:  whether to try to write a register                    */
/*          writeData: data to write to register (if any)                    */
/*                                                                           */
/*      outputs:                                                             */
/*          readData1, readData2:  data read from registers                  */
/*                                                                           */
/*     This emulates the MIPS register set semantics:  writes happen         */
/*     before reads, and we don't write to $0                                */
/*                                                                           */
/*****************************************************************************/
void registers(Data regs[NUMREGS], RegNo readReg1, RegNo readReg2, RegNo writeReg, Bool regWrite, Data writeData, Data *readData1, Data *readData2)
{

    INPUTS();
    REPORTIN(readReg1, 2);
    REPORTIN(readReg2, 2);
    REPORTIN(writeReg, 2);
    REPORTIN(regWrite, 1);
    REPORTIN(writeData, 8);

    assert(readReg1 < NUMREGS);
    assert(readReg2 < NUMREGS);
    assert(writeReg < NUMREGS);
    assert(readData1 != NULL);
    assert(readData2 != NULL);

    if (regWrite && (writeReg != 0))
        regs[writeReg] = writeData;

    *readData1 = regs[readReg1];
    *readData2 = regs[readReg2];

    OUTPUTS();
    REPORTOUT(readData1, 8);
    REPORTOUT(readData2, 8);
}

/*****************************************************************************/
/*                                                                           */
/*  signExtend                                                               */
/*                                                                           */
/*      sign extender                                                        */
/*                                                                           */
/*      input:                                                               */
/*          in:  16 bit value to sign-extend                                 */
/*                                                                           */
/*      output:                                                              */
/*          out: 32 bit sign-extended result                                 */
/*                                                                           */
/*****************************************************************************/
void signExtend(Data in, Data *out)
{
    INPUTS();
    REPORTIN(in, 4);
    assert(out != NULL);

    if (in & SIGN16)
        *out = in | HIGH16;
    else
        *out = in & LOWS16;

    OUTPUTS();
    REPORTOUT(out, 8);
}


/*****************************************************************************/
/*                                                                           */
/*  shiftLeft2                                                               */
/*                                                                           */
/*      left shifter                                                         */
/*                                                                           */
/*      input:                                                               */
/*          in:  32 bit value to be left-shifted by two bits                 */
/*                                                                           */
/*      output:                                                              */
/*          out: 32 bit left-shifted value                                   */
/*                                                                           */
/*****************************************************************************/
void shiftLeft2(Data in, Data *out)
{
    INPUTS();
    REPORTIN(in, 8);

    assert(out != NULL);

    *out = in << 2;

    OUTPUTS();
    REPORTOUT(out, 8);
}


/*****************************************************************************/
/*                                                                           */
/*  jCombine                                                                 */
/*                                                                           */
/*      jump address combiner                                                */
/*                                                                           */
/*      input:                                                               */
/*          pc:  high-order four bits of PC (+4)                             */
/*          jaddr: low order 28 bits of jump adder from instruction          */
/*                                                                           */
/*      output:                                                              */
/*          out: jump target address                                         */
/*                                                                           */
/*****************************************************************************/
void jCombine(Address pc, Address jaddr, Address *jumpAddr)
{
    INPUTS();
    REPORTIN(pc, 8);
    REPORTIN(jaddr, 8);

    assert(jumpAddr != NULL);

    *jumpAddr = pc | jaddr;

    OUTPUTS();
    REPORTOUT(jumpAddr, 8);
}

/*****************************************************************************/
/*                                                                           */
/*  mux                                                                      */
/*                                                                           */
/*      two- or three-input mux                                              */
/*                                                                           */
/*      input:                                                               */
/*          inputs:  number of inputs                                        */
/*          in0, in1, in2:  data inputs                                      */
/*          whicone: selector input                                          */
/*                                                                           */
/*      output:                                                              */
/*          out: value passed through mux                                    */
/*                                                                           */
/*                                                                           */
/*       in order to avoid duplicating  code writing several muxes, this     */
/*       is a single mux with a variable number of inputs.  There is a       */
/*       limit on the max number of inputs (set by MAXMUX), which "inputs"   */
/*       is checked against.  whichone is then checked against inputs.       */
/*                                                                           */
/*       The "do a shared mux with custom muxes calling it" (like with       */
/*       the memories) was considered, but would have had too much code in   */
/*       the derived muxes.                                                  */
/*                                                                           */
/*****************************************************************************/
void mux(unsigned int inputs, Data in0, Data in1, Data in2, MuxSel whichone, Data *out)
{
    INPUTS();
    REPORTIN(inputs, 1);
    REPORTIN(in0, 8);
    REPORTIN(in1, 8);
    REPORTIN(in2, 8);
    REPORTIN(whichone, 1);

    assert(out != NULL);
    assert(inputs <= MAXMUX);
    assert(whichone < inputs);

    switch (whichone) {
      case 0:
        *out = in0;
        break;

      case 1:
        *out = in1;
        break;

      case 2:
        *out = in2;
        break;
    }

    OUTPUTS();
    REPORTOUT(out, 8);
}
/*****************************************************************************/
/*                                                                           */
/*  aluControl                                                               */
/*                                                                           */
/*     alu control unit                                                      */
/*                                                                           */
/*     inputs:                                                               */
/*         aluOp:  aluOp from main control unit                              */
/*         func:   func field from instruction                               */
/*                                                                           */
/*     output:                                                               */
/*         aluControlOut                                                     */
/*                                                                           */
/*****************************************************************************/
void aluControl(AluOp aluOp, Func func, AluCtrl *aluControlOut)
{
    AluCtrl resFromAluOp[] = {
      ALUADD,  // 00
      ALUSUB,  // 01
      CTRLFUNC // 10
    };

    AluCtrl resFromFunc[] = {
      ALUAND, //  0
      ALUAND, //  1
      ALUAND, //  2
      ALUAND, //  3
      ALUAND, //  4
      ALUAND, //  5
      ALUAND, //  6
      ALUAND, //  7
      ALUAND, //  8
      ALUAND, //  9
      ALUAND, // 10
      ALUAND, // 11
      ALUAND, // 12
      ALUAND, // 13
      ALUAND, // 14
      ALUAND, // 15
      ALUAND, // 16
      ALUAND, // 17
      ALUAND, // 18
      ALUAND, // 19
      ALUAND, // 20
      ALUAND, // 21
      ALUAND, // 22
      ALUAND, // 23
      ALUAND, // 24
      ALUAND, // 25
      ALUAND, // 26
      ALUAND, // 27
      ALUAND, // 28
      ALUAND, // 29
      ALUAND, // 30
      ALUAND, // 31
      ALUADD, // 32 add
      ALUAND, // 33
      ALUSUB, // 34 sub
      ALUAND, // 35
      ALUAND, // 36
      ALUAND, // 37
      ALUAND, // 38
      ALUAND, // 39
      ALUAND, // 40
      ALUAND, // 41
      ALUAND, // 42
      ALUAND, // 43
      ALUAND, // 44
      ALUAND, // 45
      ALUAND, // 46
      ALUAND, // 47
      ALUAND, // 48
      ALUAND, // 49
      ALUAND, // 50
      ALUAND, // 51
      ALUAND, // 52
      ALUAND, // 53
      ALUAND, // 54
      ALUAND, // 55
      ALUAND, // 56
      ALUAND, // 57
      ALUAND, // 58
      ALUAND, // 59
      ALUAND, // 60
      ALUAND, // 61
      ALUAND, // 62
      ALUAND  // 63
    };

    INPUTS();
    REPORTIN(aluOp, 1);
    REPORTIN(func, 2);

    assert(aluOp < NUMELEMS(resFromAluOp));
    assert(func < NUMELEMS(resFromFunc));
    assert(aluControlOut != NULL);

    if (aluOp != CTRLFUNC)
        *aluControlOut = resFromAluOp[aluOp];
    else
        *aluControlOut = resFromFunc[func];

    OUTPUTS();
    REPORTOUT(aluControlOut, 2);
}

/*****************************************************************************/
/*                                                                           */
/*  and                                                                      */
/*                                                                           */
/*      and gate                                                             */
/*                                                                           */
/*      inputs:                                                              */
/*          in1, in2                                                         */
/*                                                                           */
/*      outputs:                                                             */
/*          out                                                              */
/*                                                                           */
/*****************************************************************************/
void and(Bool in1, Bool in2, Bool *out)
{
    INPUTS();
    REPORTIN(in1, 1);
    REPORTIN(in2, 1);

    assert (out != NULL);

    *out = in1 & in2;

    OUTPUTS();
    REPORTOUT(out, 1);
}

/*****************************************************************************/
/*                                                                           */
/*  pc                                                                       */
/*                                                                           */
/*      program counter register                                             */
/*                                                                           */
/*      input:                                                               */
/*          in                                                               */
/*                                                                           */
/*      outputs:                                                             */
/*          out                                                              */
/*                                                                           */
/*      returns:                                                             */
/*          TRUE if processing should continue, FALSE if the new PC has the  */
/*          same value as the old PC -- gives a way to halt the simulation   */
/*          without adding a bogus "halt" instruction                        */
/*                                                                           */
/*****************************************************************************/
void pc(Address in, Address *out)
{
    INPUTS();
    REPORTIN(in, 8);

    assert(out != NULL);

    *out = in;

    OUTPUTS();
    REPORTOUT(out, 8);

}

/*******************************************************************/
/*                                                                 */
/*  All the pipeline regs work the same:  just copy input to output*/
/*                                                                 */
/*******************************************************************/
void ifId(Instruction instrMemOut, Instruction *ifId_instrMemOut,
			 Address  pcAluOut,	   Address     *ifId_pcAluOut)
{
	INPUTS();
	REPORTIN(instrMemOut, 8);
	REPORTIN(pcAluOut, 8);

	*ifId_instrMemOut = instrMemOut;
	*ifId_pcAluOut    = pcAluOut;

	OUTPUTS();
	REPORTOUT(ifId_instrMemOut, 8);
	REPORTOUT(ifId_pcAluOut, 8);
}

void idEx(RegDstSel regDst,        RegDstSel *idEx_regDst,
		  AluOp     aluOp,		   AluOp     *idEx_aluOp,
		  AluSrcSel aluSrc,		   AluSrcSel *idEx_aluSrc,
		  Bool      branch,		   Bool      *idEx_branch,
		  Bool      memRead,	   Bool      *idEx_memRead,
		  Bool      memWrite,	   Bool      *idEx_memWrite,
		  Bool      memToReg,	   Bool      *idEx_memToReg,
		  Bool      regWrite,	   Bool      *idEx_regWrite,
		  Bool      jump,		   Bool      *idEx_jump,

		  Address   ifId_pcAluOut, Address   *idEx_pcAluOut,

		  Data      readData1,     Data      *idEx_readData1,
		  Data      readData2,     Data      *idEx_readData2,
		  Data      signExtendOut, Data      *idEx_signExtendOut,
		  Data      jAddress,      Data      *idEx_jAddress,
		  Data      zeroExtendOut, Data      *idEx_zeroExtendOut,
		  RegNo     rdField,       RegNo    *idEx_rdField,
		  RegNo     rtField,       RegNo    *idEx_rtField)
{
	INPUTS();
	REPORTIN(regDst, 1);
	REPORTIN(aluOp, 2);
	REPORTIN(aluSrc, 1);
	REPORTIN(branch, 1);
	REPORTIN(memRead, 1);
	REPORTIN(memWrite, 1);
	REPORTIN(memToReg, 1);
	REPORTIN(regWrite, 1);
	REPORTIN(jump, 1);

	REPORTIN(ifId_pcAluOut, 8);

	REPORTIN(readData1, 8);
	REPORTIN(readData2, 8);
	REPORTIN(signExtendOut, 8);
	REPORTIN(jAddress, 8);
	REPORTIN(zeroExtendOut, 8);
	REPORTIN(rdField, 2);
	REPORTIN(rtField, 2);

	*idEx_regDst        = regDst;
	*idEx_aluOp         = aluOp;
	*idEx_aluSrc        = aluSrc;
	*idEx_branch        = branch;
	*idEx_memRead       = memRead;
	*idEx_memWrite      = memWrite;
	*idEx_memToReg      = memToReg;
	*idEx_regWrite      = regWrite;
	*idEx_jump          = jump;

	*idEx_pcAluOut      = ifId_pcAluOut;

	*idEx_readData1     = readData1;
	*idEx_readData2     = readData2;
	*idEx_signExtendOut = signExtendOut;
	*idEx_jAddress      = jAddress;
	*idEx_zeroExtendOut = zeroExtendOut;
	*idEx_rdField       = rdField;
	*idEx_rtField       = rtField;

	OUTPUTS();
	REPORTOUT(idEx_regDst, 1);
	REPORTOUT(idEx_aluOp, 1);
	REPORTOUT(idEx_aluSrc, 1);
	REPORTOUT(idEx_branch, 1);
	REPORTOUT(idEx_memRead, 1);
	REPORTOUT(idEx_memWrite, 1);
	REPORTOUT(idEx_memToReg, 1);
	REPORTOUT(idEx_regWrite, 1);
	REPORTOUT(idEx_jump, 1);

	REPORTOUT(idEx_pcAluOut, 8);

	REPORTOUT(idEx_readData1, 8);
	REPORTOUT(idEx_readData2, 8);
	REPORTOUT(idEx_signExtendOut, 8);
	REPORTOUT(idEx_jAddress, 8);
	REPORTOUT(idEx_zeroExtendOut, 8);
	REPORTOUT(idEx_rdField, 2);
	REPORTOUT(idEx_rtField, 2);
}

void exMem(Bool    idEx_branch,    Bool    *exMem_branch,
		   Bool    idEx_jump,      Bool    *exMem_jump,
		   Bool    idEx_memRead,   Bool    *exMem_memRead,
		   Bool    idEx_memWrite,  Bool    *exMem_memWrite,
		   Bool    idEx_memToReg,  Bool    *exMem_memToReg,
		   Bool    idEx_regWrite,  Bool    *exMem_regWrite,
		   
		   Address jCombineOut,       Address *exMem_jCombineOut,
		   Address branchAddrOut,  Address *exMem_branchAddrOut,
		   
		   Bool    zero,           Bool    *exMem_zero,
		   Data    aluResult,      Data    *exMem_aluResult,
		   Data    idEx_readData2, Data    *exMem_readData2,
		   RegNo   writeRegMuxOut, RegNo  *exMem_writeRegMuxOut)
{
	INPUTS();
	REPORTIN(idEx_branch, 1);
	REPORTIN(idEx_jump, 1);
	REPORTIN(idEx_memRead, 1);
	REPORTIN(idEx_memWrite, 1);
	REPORTIN(idEx_memToReg, 1);
	REPORTIN(idEx_regWrite, 1);

	REPORTIN(jCombineOut, 8);
	REPORTIN(branchAddrOut, 8);

	REPORTIN(zero, 1);
	REPORTIN(aluResult, 8);
	REPORTIN(idEx_readData2, 8);
	REPORTIN(writeRegMuxOut, 2);

	*exMem_branch         = idEx_branch;
	*exMem_jump           = idEx_jump;
	*exMem_memRead        = idEx_memRead;
	*exMem_memWrite       = idEx_memWrite;
	*exMem_memToReg       = idEx_memToReg;
	*exMem_regWrite       = idEx_regWrite;

	*exMem_jCombineOut    = jCombineOut;
	*exMem_branchAddrOut  = branchAddrOut;

	*exMem_zero           = zero;
	*exMem_aluResult      = aluResult;
	*exMem_readData2      = idEx_readData2;
	*exMem_writeRegMuxOut = writeRegMuxOut;

	OUTPUTS();
	REPORTOUT(exMem_branch, 1);
	REPORTOUT(exMem_jump, 1);
	REPORTOUT(exMem_memRead, 1);
	REPORTOUT(exMem_memWrite, 1);
	REPORTOUT(exMem_memToReg, 1);
	REPORTOUT(exMem_regWrite, 1);

	REPORTOUT(exMem_jCombineOut, 8);
	REPORTOUT(exMem_branchAddrOut, 8);

	REPORTOUT(exMem_zero, 1);
	REPORTOUT(exMem_aluResult, 8);
	REPORTOUT(exMem_readData2, 8);
	REPORTOUT(exMem_writeRegMuxOut, 2);
}

void memWb(Bool   exMem_memToReg,       Bool   *memWb_memToReg,
		   Bool   exMem_regWrite,       Bool   *memWb_regWrite,
		   RegNo  exMem_writeRegMuxOut, RegNo *memWb_writeRegMuxOut,

		   Data   readData,             Data   *memWb_readData,
		   Data   exMem_aluResult,      Data   *memWb_aluResult)
{
	INPUTS();
	REPORTIN(exMem_memToReg, 1);
	REPORTIN(exMem_regWrite, 1);
	REPORTIN(exMem_writeRegMuxOut, 2);

	REPORTIN(readData, 8);
	REPORTIN(exMem_aluResult, 8);

	*memWb_memToReg       = exMem_memToReg;
	*memWb_regWrite       = exMem_regWrite;
	*memWb_writeRegMuxOut = exMem_writeRegMuxOut;

	*memWb_readData       = readData;
	*memWb_aluResult      = exMem_aluResult;

	OUTPUTS();
	REPORTOUT(memWb_memToReg, 1);
	REPORTOUT(memWb_regWrite, 1);
	REPORTOUT(memWb_writeRegMuxOut, 2);

	REPORTOUT(memWb_readData, 8);
	REPORTOUT(memWb_aluResult, 8);
}
